# matrix-appservice-kakaotalk - A Matrix-KakaoTalk puppeting bridge.
# Copyright (C) 2022 Tulir Asokan, Andrew Ferrazzutti
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, List, Set

from asyncpg import Record
from attr import dataclass

from mautrix.types import RoomID, UserID
from mautrix.util.async_db import Database

from ..kt.types.bson import Long

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class User:
    db: ClassVar[Database] = fake_db

    mxid: UserID
    ktid: Long | None
    uuid: str | None
    access_token: str | None
    refresh_token: str | None
    notice_room: RoomID | None

    @classmethod
    def _from_row(cls, row: Record) -> User:
        data = {**row}
        ktid = data.pop("ktid", None)
        return cls(**data, ktid=Long.from_optional_bytes(ktid))

    @classmethod
    def _from_optional_row(cls, row: Record | None) -> User | None:
        return cls._from_row(row) if row is not None else None

    @classmethod
    async def all_logged_in(cls) -> List[User]:
        q = """
            SELECT mxid, ktid, uuid, access_token, refresh_token, notice_room FROM "user"
            WHERE ktid<>0
        """
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows if row]

    @classmethod
    async def get_by_ktid(cls, ktid: Long) -> User | None:
        q = 'SELECT mxid, ktid, uuid, access_token, refresh_token, notice_room FROM "user" WHERE ktid=$1'
        row = await cls.db.fetchrow(q, bytes(ktid))
        return cls._from_optional_row(row)

    @classmethod
    async def get_by_mxid(cls, mxid: UserID) -> User | None:
        q = 'SELECT mxid, ktid, uuid, access_token, refresh_token, notice_room FROM "user" WHERE mxid=$1'
        row = await cls.db.fetchrow(q, mxid)
        return cls._from_optional_row(row)

    @classmethod
    async def get_all_uuids(cls) -> Set[str]:
        q = 'SELECT uuid FROM "user" WHERE uuid IS NOT NULL'
        return {tuple(record)[0] for record in await cls.db.fetch(q)}

    async def insert(self) -> None:
        q = """
            INSERT INTO "user" (mxid, ktid, uuid, access_token, refresh_token, notice_room) 
            VALUES ($1, $2, $3, $4, $5, $6)
        """
        await self.db.execute(
            q, self.mxid, Long.to_optional_bytes(self.ktid), self.uuid, self.access_token, self.refresh_token, self.notice_room
        )

    async def delete(self) -> None:
        await self.db.execute('DELETE FROM "user" WHERE mxid=$1', self.mxid)

    async def save(self) -> None:
        q = """
            UPDATE "user" SET ktid=$1, uuid=$2, access_token=$3, refresh_token=$4, notice_room=$5 
            WHERE mxid=$6
        """
        await self.db.execute(
            q, Long.to_optional_bytes(self.ktid), self.uuid, self.access_token, self.refresh_token, self.notice_room, self.mxid
        )
